module Polarity where


-- polarities and strict positivity ----------------------------------

data Pos 
  = SPos  -- strictly positive 
  | Pos   -- positive
  | Mixed -- unknown polarity (invariant)
  | Neg   -- negative, used internally for contravariance of sized codata  
  | NOcc  -- non-occurring, used to determine whether data type is recursive
         deriving (Eq,Ord)

instance Show Pos where
  show SPos  = "++"
  show Pos   = "+"
  show Mixed = "o"
  show Neg   = "-"
  show NOcc  = "/"

-- information ordering
leqPol :: Pos -> Pos -> Bool
leqPol x NOcc   = True   -- NOcc is top
leqPol NOcc x   = False
leqPol Mixed y  = True   -- Mixed is bottom
leqPol x Mixed  = False
leqPol Pos SPos = True 
leqPol x   y    = x == y

isSPos :: Pos -> Bool
isSPos SPos = True
isSPos NOcc = True
isSPos _ = False

isPos :: Pos -> Bool
isPos Pos = True
isPos x = isSPos x

polNeg :: Pos -> Pos
polNeg SPos  = Neg
polNeg Pos   = Neg
polNeg Neg   = Pos
polNeg Mixed = Mixed
polNeg NOcc  = NOcc

-- polarity composition
polComp :: Pos -> Pos -> Pos
polComp NOcc  x   = NOcc   -- most dominant
polComp x NOcc    = NOcc
polComp Mixed x   = Mixed  -- dominant except for NOcc
polComp x Mixed   = Mixed
polComp SPos  x   = x      -- neutral
polComp x SPos    = x
polComp Pos  x    = x      -- neutral except for SPos
polComp x Pos     = x
polComp Neg Neg   = Pos    -- order 2

-- polarity inverse composition (see Abel, MSCS 2008)
-- invComp p q1 <= q2  <==> q1 <= polComp p q2 
invComp :: Pos -> Pos -> Pos
invComp Mixed Mixed = Mixed
invComp Mixed x     = NOcc
invComp NOcc  x     = Mixed
invComp SPos  x     = x         -- SPos is the identity
invComp p     SPos  = NOcc      -- SPos preserved only under SPos
invComp Pos   x     = x         -- x not SPos
invComp Neg   x     = polNeg x  -- x not SPos

-- polarity conjuction (infimum)
polAnd :: Pos -> Pos -> Pos
polAnd NOcc x       = x       -- most information
polAnd x NOcc       = x
polAnd x y | x == y = x       -- same information
polAnd SPos Pos     = Pos     -- SPos is more informative than Pos
polAnd Pos SPos     = Pos
polAnd _ _          = Mixed   -- remaining cases: conflicting or no info
{-
polAnd Mixed x  = Mixed   -- least information
polAnd x Mixed  = Mixed
polAnd SPos Neg = Mixed
polAnd Neg SPos = Mixed
-}

relPol :: Pos -> (a -> a -> Bool) -> (a -> a -> Bool)
relPol NOcc  r a b = True
relPol Mixed r a b = r a b && r b a
relPol Neg   r a b = r b a
relPol Pos   r a b = r a b
relPol SPos  r a b = r a b

relPolM :: (Monad m) => Pos -> (a -> a -> m ()) -> (a -> a -> m ())
relPolM NOcc  r a b = return ()
relPolM Mixed r a b = r a b >> r b a
relPolM Neg   r a b = r b a
relPolM Pos   r a b = r a b
relPolM SPos  r a b = r a b
