{- A data structure to represent a forest of upside down trees,
similar to union-find.  The idea is to manage a tree-shaped form of
strict inequations

  i1 > i2 > i3
     > j2 > j3 > j4 > j5
          > k3
          > l3 > l4

  m1 > m2
  
  n1

Checking inequalty x < y is then performed by just enumerating the
parents of x and checking wether y is a member of it.
-}

module TreeShapedOrder where

import Data.List hiding (insert) -- groupBy

import Data.Map (Map)
import qualified Data.Map as Map

import Data.Tree (Tree(..), Forest) -- rose trees
import qualified Data.Tree as Tree

import Util -- headM

-- | Tree-structured partial orders.
newtype TSO a = TSO { unTSO :: Map a a } deriving (Eq, Ord)

-- | Empty TSO.
empty :: TSO a
empty = TSO $ Map.empty

-- | @insert a b o@  inserts a with parent b into order o.
-- It does not check whether the tree structure is preserved. 
insert :: (Ord a, Eq a) => a -> a -> TSO a -> TSO a
insert a b (TSO o) = TSO $ Map.insert a b o

-- | Construction for a list of child-parent pairs.
fromList :: (Ord a, Eq a) => [(a,a)] -> TSO a
fromList l = foldl (\ o (a,b) -> insert a b o) empty l

-- | @parents a0 o = [a1,..,an]@ lists the parents of @a0@ in order,
-- i.e., a(i+1) is parent of a(i).
parents :: (Ord a, Eq a) => a -> TSO a -> [a]
parents a (TSO o) = loop (Map.lookup a o) where
  loop Nothing  = []
  loop (Just b) = b : loop (Map.lookup b o)

-- | @parent a o@ returns the immediate parent, if it exists.
parent :: (Ord a, Eq a) => a -> TSO a -> Maybe a
parent a t = headM $ parents a t 

-- | @isAncestor a b o = Just n@ if there are n steps up from a to b.  
isAncestor :: (Ord a, Eq a) => a -> a -> TSO a -> Maybe Int
isAncestor a b o = elemIndex b (a : parents a o)

-- | @diff a b o = Just k@ if there are k steps up from a to b  
-- or (-k) steps down from b to a.
diff ::  (Ord a, Eq a) => a -> a -> TSO a -> Maybe Int
diff a b o = maybe (fmap (\ k -> -k) $ isAncestor b a o) Just $ isAncestor a b o

-- | create a map from parents to list of sons, leaves have an empty list  
invert :: (Ord a, Eq a) => TSO a -> Map a [a]
invert (TSO o) = Map.foldWithKey step Map.empty o where
  step son parent m = Map.insertWith (++) son [] $ 
    Map.insertWith (++) parent [son] m

-- | get the leaves of the TSO forest
leaves :: (Ord a, Eq a) => TSO a -> [a]
leaves o = map fst $ filter (\ (k,a) -> null a) $ Map.toList (invert o)

{- FLAWED BOTTOM-UP-ATTEMPT, DOES NOT WORK
{- How to invert a TSO?

1. Create a Map from parents to their list of children.

2. Keep a working set of nodes.  
   Find the leafs in this working set (nodes that do not have children).
   Cluster them by their parents.
   Turn their parents into trees,
   Continue with the parents.
-}
-- | invert a tree shaped order into a forest.  This can be used for printing
toForest :: (Ord a, Eq a) => TSO a -> Forest a
toForest o = loop (step initialTrees) where 
  initialTrees = map (flip Node []) $ leaves o
  -- step :: (Ord a, Eq a) => Forest a -> [(Maybe a, Forest a)]
  step ts = map (\ l -> (fst (head l), map snd l)) $
    groupBy (\ (p,t) (p',t') -> p == p') $ 
    sortBy (\ (p,t) (p',t') -> compare p p') $ 
    map (\ t -> (parent (rootLabel t) o, t)) ts 
  -- loop :: (Ord a, Eq a) => [(Maybe a, Forest a)] -> Forest a
  loop [] = []
  -- the trees whose roots have no parents are parts of the final forest 
  loop ((Nothing, roots) : nonroots) = roots ++ loop nonroots
  -- the trees whose roots have a parent are iterated
  loop nonroots = loop $ step $ map (\ (Just p, ts) -> Node p ts) nonroots
-}

-- take a lexicographically sorted list of pathes
-- and turn it into a forest by
-- gathering the lists by common prefixes
pathesToForest :: (Ord a, Eq a) => [[a]] -> Forest a
pathesToForest [] = []
pathesToForest ll =
  map (\ l -> Node (head (head l)) 
                   (pathesToForest $ filter (not . null) $ map tail l)) $ 
    groupBy (\ l l' -> head l == head l') ll

-- | invert a tree shaped order into a forest.  This can be used for printing.
toForest :: (Ord a, Eq a) => TSO a -> Forest a
toForest o = pathesToForest $ sort $ map (\ a -> reverse (a : parents a o)) $ leaves o -- lex. sort

instance (Ord a, Eq a, Show a) => Show (TSO a) where
  show o = Tree.drawForest $ map (fmap show) $ toForest o 

{-
draw :: (Ord a, Eq a, Show a) => TSO a -> String
draw o = Tree.drawForest $ map (fmap show) $ toForest o 
-}

-- test

l1 = map (\ (k,l) -> ("i" ++ show k, "i" ++ show l)) [(0,1),(1,2),(2,3),(3,4)]
     ++ [("j2","i3")]
o1 = fromList l1
t1 = diff "i2" "i1" o1
t2 = diff "i2" "j2" o1
